@@ -1,4 +1,4 @@
-import pyrox.http.filtering as filtering
+import pyrox.filtering as filtering
 
 
 class RoutingFilter(filtering.HttpFilter):
@@ -6,5 +6,7 @@ class RoutingFilter(filtering.HttpFilter):
     This is an example on how to specify an upstream route when a request is
     intercepted.
     """
-    def on_request(self, request_message):
+
+    @filtering.handles_request_head
+    def on_request_head(self, request_message):
         return filtering.route('google.com:80')
