@@ -18,17 +18,30 @@ _MAX_READ = 1024
 _CHUNK_CLOSE = b'0\r\n\r\n'
 
 
-def _write_to_stream(stream, bytes, is_chunked):
+def _write_to_stream(stream, data, is_chunked):
     if is_chunked:
         # Format and write this chunk
         chunk = bytearray()
-        chunk.extend(hex(length)[2:])
+        chunk.extend(hex(len(data))[2:])
         chunk.extend('\r\n')
-        chunk.extend(bytes)
+        chunk.extend(data)
         chunk.extend('\r\n')
-        stream.write(chunk)
+
+        stream.write(bytes(chunk))
     else:
-        stream.write(bytes)
+        stream.write(data)
+
+
+class AccumulationStream(object):
+
+    def __init__(self):
+        self.bytes = bytearray()
+
+    def write(self, data):
+        self.bytes.extend(data)
+
+    def size(self):
+        return len(self.bytes)
 
 
 class ProxyHandler(ParserDelegate):
@@ -63,6 +76,7 @@ class DownstreamHandler(ProxyHandler):
     This data comes from the client initiating the request against the
     proxy.
     """
+
     def __init__(self, downstream, filter_pl, connect_upstream):
         super(DownstreamHandler, self).__init__(filter_pl, HttpRequest())
         self._downstream = downstream
@@ -92,7 +106,12 @@ class DownstreamHandler(ProxyHandler):
 
     def on_headers_complete(self):
         # Execute against the pipeline
-        action = self._filter_pl.on_request(self._http_msg)
+        action = self._filter_pl.on_request_head(self._http_msg)
+
+        self._http_msg.remove_header('content-length')
+        self._http_msg.remove_header('transfer-encoding')
+
+        self._http_msg.header('transfer-encoding').values.append('chunked')
 
         # If we're rejecting then we're not going to connect to upstream
         if action.is_rejecting():
@@ -108,13 +127,24 @@ class DownstreamHandler(ProxyHandler):
     def on_body(self, bytes, length, is_chunked):
         # Rejections simply discard the body
         if not self._rejected:
+            accumulator = AccumulationStream()
+            data = bytes[:length]
+
+            print(data)
+
+            self._filter_pl.on_request_body(data, accumulator)
+
+            if accumulator.size() > 0:
+                print('here')
+                data = accumulator.bytes
+
             # If we're not already connected, store the fragment for later
             if not self._upstream:
-                self._store_body(bytes)
+                self._store_body(data)
             else:
                 _write_to_stream(
                     self._upstream,
-                    self._get_body(bytes),
+                    self._get_body(data),
                     is_chunked)
 
     def on_message_complete(self, is_chunked, keep_alive):
@@ -134,27 +164,41 @@ class UpstreamHandler(ProxyHandler):
     data usually comes from the origin service or it may come from another
     proxy.
     """
+
     def __init__(self, downstream, upstream, filter_pl):
         super(UpstreamHandler, self).__init__(filter_pl, HttpResponse())
         self._downstream = downstream
         self._upstream = upstream
 
     def on_status(self, status_code):
-        self._http_msg.status_code = str(status_code)
+        self._http_msg.status = str(status_code)
 
     def on_headers_complete(self):
-        action = self._filter_pl.on_response(self._http_msg)
+        action = self._filter_pl.on_response_head(self._http_msg)
+
+        self._http_msg.remove_header('content-length')
+        self._http_msg.remove_header('transfer-encoding')
+
+        self._http_msg.header('transfer-encoding').values.append('chunked')
+
         if action.is_rejecting():
             self._rejected = True
-            self._response = action.response
+            self._response = action.payload
         else:
             self._downstream.write(self._http_msg.to_bytes())
 
     def on_body(self, bytes, length, is_chunked):
         # Rejections simply discard the body
-        if self._rejected:
-            return
-        _write_to_stream(self._downstream, bytes, is_chunked)
+        if not self._rejected:
+            accumulator = AccumulationStream()
+            data = bytes[:length]
+
+            self._filter_pl.on_response_body(data, accumulator)
+
+            if accumulator.size() > 0:
+                data = accumulator.bytes
+
+            _write_to_stream(self._downstream, data, True)
 
     def on_message_complete(self, is_chunked, keep_alive):
         callback = None if keep_alive else self._downstream.close
@@ -163,7 +207,7 @@ class UpstreamHandler(ProxyHandler):
             self._downstream.write(
                 self._http_msg.to_bytes(),
                 callback=callback)
-        elif is_chunked:
+        else:
             # Finish the last chunk.
             self._downstream.write(
                 _CHUNK_CLOSE,
